/*
 MPI/cudaMPI/glMPI Communications Benchmark Program:
Computes the short-message ping time, long message bandwidth, etc.

Orion Sky Lawlor, olawlor@acm.org, 4/30/2002
*/

#include <mpi.h> /* for MPI_ functions */
#include <cuda_runtime.h>

#include <stdio.h> /* for printf */
#include <stdlib.h> /* for atexit */
#include <math.h> /* for "cos" */
#include <sys/stat.h> /* for "mkdir" */
#include <vector>
#include <algorithm> /* for "pair" */


#include "ogl/gpgpu.h"
#include "ogl/glsl.cpp" /* avoid linker hassles, by including here */
#include "glMPI.h"

int doGLfinish=0;


#include "cudaMPI.cpp" /* for cudaMPI_ functions */

int doCudaSync=0;
#include "inc.cpp"  /* for timing functions */

/*Number of iterations for "short" operations (like send)*/
#define SHORTITER 1000
/*Number of iterations for "long" operations (like reduce)*/
#define LONGITER 1000
/*Number of bytes for "long" message*/
#define LONGMSG 1024*1024

#define STARTTIME(str) \
	curtest=str;\
	if (0 && mype==0) printf("Starting test %s\n",curtest);\
	MPI_Barrier(MPI_COMM_WORLD); \
	n=0; \
	starttime=MPI_Wtime();\

#define ENDTIME \
	starttime=(MPI_Wtime()-starttime)/n; \
	if (mype==0) {\
		printf("Test %s took %.3fus per iteration (%d iterations)\n",\
			curtest,1.0e6*starttime,n);\
		printf("%d\t%.9f\t%d\tGREP_%s\n\n",totpe,starttime,n,curtest);\
	}\

double clock_overhead=0.0;
void test_clock(void) {
	double t,q;
	enum {n=10};
	double tries[n];
	for (int i=0;i<n;i++) {
		t=MPI_Wtime();
		while (t==(q=MPI_Wtime())) {}
		tries[i]=q-t; /* clock quantization */
	}
	std::sort(&tries[0],&tries[n]);
	double clock_quant=tries[n/2];
	printf("Clock quantization: %.1f ns\n",1.0e9*clock_quant);
	
	enum {m=1000};
	t=MPI_Wtime();
	int count=1;
	while (MPI_Wtime()<t+m*clock_quant) count++; /* wait for 1000 quantization periods */
	clock_overhead=(MPI_Wtime()-t)/count;
	printf("Clock overhead: %.1f ns\n",1.0e9*clock_overhead);
}


double fear_me_compiler=0.0;

/**
  Spend this many microseconds doing (useless) computation on the CPU.
  
  Must be calibrated for each CPU & compiler flag combination
*/
int sandbag(int microsec) {
	int m,j;
	double sum=0.0;
	for (m=0;m<microsec;m++) {
		for (j=0;j<18;j++)
			sum+=cos(j*microsec);
	}
	fear_me_compiler+=sum;
	return (int)sum;
}


/************ Validate linear performance model for this function */
std::string log_dir="log"; /* where output data is logged */
std::string log_suffix=""; /* suffix typically == the recording node number */
int mype=0,totpe=1; /* MPI ranks */
float step_pow=2.0; /* take steps of size this power of two */
int verbose=10;

void debug(std::string what) {
	if (mype!=-1) return; // not the bad PE
	
	static FILE *f=fopen("debug_log.txt","w");
	fprintf(f,"%d %s\n",mype,what.c_str()); fflush(f);
}


/* Function to switch timing sizes to "new_size".  Actual size is returned. */
typedef int (*switchsize_fn)(int new_size);

enum {max_size=16*1024*1024};

void validate(std::string run_name,timeable_fn fn,switchsize_fn sw,
	int first_size=1,int last_size=max_size)
{
	debug("starting test "+run_name);
	FILE *f=0;
	if (mype==0) f=fopen((log_dir+"/"+run_name+log_suffix).c_str(),"w");
	std::vector<std::pair<int,double> > runs; /* size, and nanoseconds */

/* Do timing runs */
	int size;
	for (size=first_size;
		size<=last_size;)
	{
		size=sw(size); /* set size */
		debug("running test");
		double t=1.0e9*time_function(fn);
		debug("saving test results");
		
		if (verbose>20 && mype==0)
		{
		    printf("%20s size=%d	%.3f ns/elt	(%.1f ns/call) \n",
			run_name.c_str(),size,t/size,t); fflush(stdout);
		}
		if (f) {fprintf(f,"%d	%.3f	%.3f\n",size,t/size,t); fflush(f);}
		
		runs.push_back(std::make_pair(size,t));
	
		/* Increment size for next time around */
		int nsize=(int)(size*std::pow(2.0f,step_pow));
		if (size==nsize) size=size+1; else size=nsize;
	}

	debug("analyzing test results");
/* Look for anomalous timings */
	int last=runs.size()-1;
	double beta=runs[last].second/runs[last].first; /* ns/elt */
	double alpha=runs[0].second-beta*runs[0].first; /* ns startup time */
	if (mype==0) {
		printf("%20s  t = %.1f ns + %.3f ns/elt",run_name.c_str(),alpha,beta);
		for (int i=0;i<=last;i++) {
			size=runs[i].first;
			double model=alpha+beta*size;
			double actual=runs[i].second;
			double err=100.0*(model-actual)/actual;
			if (fabs(err)>50) 
				printf("\n%20sBUT	size=%d	 model=%.3f ns actual=%.1fns  (%.0f%% error)",
					"",size,model,actual,err);
			fflush(stdout);
		}
		printf("\n");
	}
	
	if (f) fclose(f);
	
	debug("done with test "+run_name);
}

/****************************************** MPI benchmarks ****/

int nMPIbytes=0;
std::vector<char> MPIdata; /* always nMPIbytes long */
std::vector<char> MPIdata2; /* always nMPIbytes long */
int switchsizeMPI(int newbytes) {
	/* Avoid test-to-test interaction */
	MPI_Barrier(MPI_COMM_WORLD);
	
	nMPIbytes=newbytes;
	MPIdata.resize(newbytes);
	MPIdata2.resize(newbytes);
	return newbytes;
}

int sendPE=0,recvPE=1;
int MPIbarrier(void) {
	MPI_Barrier(MPI_COMM_WORLD);
	return 0;
}
int MPIbcast(void) {
	MPI_Bcast(&MPIdata[0],nMPIbytes,MPI_BYTE, 0,MPI_COMM_WORLD);
	MPI_Barrier(MPI_COMM_WORLD);
	return 0;
}
int MPIreduce(void) {
	MPI_Reduce(&MPIdata[0],&MPIdata2[0],nMPIbytes/4,MPI_INT,MPI_SUM, 0,MPI_COMM_WORLD);
	MPI_Barrier(MPI_COMM_WORLD);
	return 0;
}
int MPIallreduce(void) {
	MPI_Allreduce(&MPIdata[0],&MPIdata2[0],nMPIbytes/4,MPI_INT,MPI_SUM,MPI_COMM_WORLD);
	return 0;
}
int MPIoneway(void) {
	if (mype==sendPE) 
		MPI_Send(&MPIdata[0],nMPIbytes,MPI_BYTE, recvPE,0,MPI_COMM_WORLD);
	MPI_Status sts;
	if (mype==recvPE) 
		MPI_Recv(&MPIdata2[0],nMPIbytes,MPI_BYTE, sendPE,0,MPI_COMM_WORLD,&sts); 
	return 0;
}
int MPIpingpong(void) {
	MPIoneway(); std::swap(sendPE,recvPE); return 0;
}

void MPI_waitreq(MPI_Request *req)
{
	MPI_Status sts;
	if (1) { /* polling */
		while (1) {
			int flag=0;
			MPI_Test(req,&flag,&sts);
			if (flag) break;
		}
	} 
	/* blocking */
	MPI_Wait(req,&sts);
}

int MPI_Ioneway(void) {
	MPI_Request req;
	if (mype==sendPE) {
		MPI_Isend(&MPIdata[0],nMPIbytes,MPI_BYTE, recvPE,0,MPI_COMM_WORLD,&req);
		MPI_waitreq(&req);
	}
	if (mype==recvPE) {
		MPI_Irecv(&MPIdata2[0],nMPIbytes,MPI_BYTE, sendPE,0,MPI_COMM_WORLD,&req);
		MPI_waitreq(&req);
	} 
	return 0;
}
void validateMPI(void) {
	validate("MPIbarrier",MPIbcast,switchsizeMPI,1,1);
	validate("MPIbcast",MPIbcast,switchsizeMPI);
	validate("MPIreduce",MPIreduce,switchsizeMPI,4);
	validate("MPIallreduce",MPIallreduce,switchsizeMPI,4);
	
	if (totpe>=2) {
		validate("MPIoneway",MPIoneway,switchsizeMPI);
		validate("MPIpingpong",MPIpingpong,switchsizeMPI);
		validate("MPI_Ioneway",MPI_Ioneway,switchsizeMPI);
	}
}



/****************************** CUDA performance model */  
int nCUDAbytes=0; /* size of below buffers */
void *CUDAhost=0; /* host buffer */
void *CUDAdevice=0; /* device buffer 1*/
void *CUDAdevice2=0; /* device buffer 2*/

void CUDAerrorcheck(const char *where) {
	/* Check for any errors */
	cudaError_t e=cudaGetLastError();
	if (e!=cudaSuccess) {
		printf("CUDA error %d on pe %d %s: %s\n",
			e,mype,where,cudaGetErrorString(e));
		exit(1);
	}
}

int CUDAswitchsize(int newbytes) {
	CUDAerrorcheck("on switchsize");

	/* Avoid test-to-test contamination */
	cudaThreadSynchronize();
	
	newbytes=(newbytes+3)&~3; /* round up to multiple of 4 */
		
	nCUDAbytes=newbytes;
	return newbytes;
}

int CUDAhostnewdel(void) {
	delete[] new unsigned char[nCUDAbytes];
	return 0;
}
int CUDAdevmalloc(void) {
	cudaMalloc(&CUDAdevice,nCUDAbytes);
	cudaFree(CUDAdevice);
	return 0;
}
int CUDAhostmalloc(void) {
	cudaMallocHost(&CUDAhost,nCUDAbytes);
	cudaFreeHost(CUDAhost);
	return 0;
}
int CUDAhostmappedmalloc(void) {
	cudaHostAlloc(&CUDAhost,nCUDAbytes,cudaHostAllocMapped+cudaHostAllocWriteCombined);
	void *devhost=0; /* host memory's device-mapped address */
	cudaHostGetDevicePointer(&devhost,CUDAhost,0);
	cudaFreeHost(CUDAhost);
	return 0;
}

int CUDAd2d(void) {
	cudaMemcpy(CUDAdevice2,CUDAdevice,nCUDAbytes,cudaMemcpyDeviceToDevice);
	return 0;
}
int CUDAh2d(void) {
	cudaMemcpy(CUDAdevice,CUDAhost,nCUDAbytes,cudaMemcpyHostToDevice);
	return 0;
}
int CUDAd2h(void) {
	cudaMemcpy(CUDAhost,CUDAdevice,nCUDAbytes,cudaMemcpyDeviceToHost);
	return 0;
}

void CUDAstreamsync(void) {
	cudaStream_t s;
	cudaStreamCreate(&s);
	cudaStreamSynchronize(s);
}

double worktime=50.0e-6;
int CUDAasyncprint=0;

void CUDAasyncCopy(void *to,void *from,int bytes,cudaMemcpyKind dir) {
	if (0) { /* create stream on the fly: 80us! */
		cudaStream_t s;
		cudaStreamCreate(&s);
		cudaMemcpyAsync(to,from,bytes,dir,s);
		cudaStreamSynchronize(s);
	}
	/* Else recycle an existing stream object */
	static int streamSetup=0;
	static cudaStream_t s;
	if (!streamSetup) {
		streamSetup=1;
		cudaStreamCreate(&s);
	}
	double start=MPI_Wtime();
	cudaMemcpyAsync(to,from,bytes,dir,s);
	double mid=MPI_Wtime();
	int workcount=0, loopcount=0;
	if (1) { /* probe-style: 47us alpha cost */
		while (1) {
			double t=MPI_Wtime();
			while (MPI_Wtime()<t+worktime) {loopcount++;} /* do "work" for so long */
			workcount++;
			/* Check the stream every so often */
			if (cudaErrorNotReady!=cudaStreamQuery(s)) break;
		}
	} else { /* blocking style: 40us alpha cost */
		cudaStreamSynchronize(s);
	}
	double e=MPI_Wtime();
	if (CUDAasyncprint) { 
		printf("cudaMemcpyAsync(%d,%.1f): %.1fus start, %.1fus work (%d), %.1fus query, %.1fus tot\n",
			bytes,1.0e6*worktime,1.0e6*(mid-start),
			1.0e6*workcount*worktime,loopcount,1.0e6*(e-mid-workcount*worktime),
			1.0e6*(e-start));
	}
}

int CUDAh2dA(void) {
	CUDAasyncCopy(CUDAdevice,CUDAhost,nCUDAbytes,cudaMemcpyHostToDevice);
	return 0;
}
int CUDAd2hA(void) {
	CUDAasyncCopy(CUDAhost,CUDAdevice,nCUDAbytes,cudaMemcpyDeviceToHost);
	return 0;
}

int cudaMPIoneway(void) {
	if (mype==sendPE) {
		cudaMPI_Send(CUDAdevice2,nCUDAbytes,MPI_BYTE, recvPE,0,MPI_COMM_WORLD);
	}
	MPI_Status sts;
	if (mype==recvPE) {
		cudaMPI_Recv(CUDAdevice,nCUDAbytes,MPI_BYTE, sendPE,0,MPI_COMM_WORLD,&sts); 
	}
	return 0;
}
int cudaMPI_Ioneway(void) {
	cudaMPI_Request req;
	MPI_Status sts;
	if (mype==sendPE) {
		cudaMPI_Isend(CUDAdevice2,nCUDAbytes,MPI_BYTE, recvPE,0,MPI_COMM_WORLD,&req);
		cudaMPI_Wait(&req,&sts);
	}
	if (mype==recvPE) {
		cudaMPI_Irecv(CUDAdevice,nCUDAbytes,MPI_BYTE, sendPE,0,MPI_COMM_WORLD,&req); 
		cudaMPI_Wait(&req,&sts);
	}
	return 0;
}
int cudaMPIpingpong(void) {
	cudaMPIoneway(); std::swap(sendPE,recvPE); return 0;
}
int cudaMPIbcast(void) {
	cudaMPI_Bcast(CUDAdevice,nCUDAbytes,MPI_BYTE, 0,MPI_COMM_WORLD);
	MPI_Barrier(MPI_COMM_WORLD);
	return 0;
}
int cudaMPIreduce(void) {
	cudaMPI_Reduce(CUDAdevice,CUDAdevice2,nCUDAbytes/4,MPI_INT,MPI_SUM, 0,MPI_COMM_WORLD);
	MPI_Barrier(MPI_COMM_WORLD);
	return 0;
}
int cudaMPIallreduce(void) {
	cudaMPI_Allreduce(CUDAdevice,CUDAdevice2,nCUDAbytes/4,MPI_INT,MPI_SUM,MPI_COMM_WORLD);
	return 0;
}


__global__ void CUDAwritezero(float *dest) {
	int i=blockDim.x*blockIdx.x+threadIdx.x;
	dest[i]=0.0;
}
int blockSize=32;
int CUDAbasicwrite(void) {
	CUDAwritezero<<<nCUDAbytes/blockSize/4,blockSize>>>((float *)CUDAdevice);
	return 0;
}

__global__ void CUDAcopykernel(const float *src,float *dest) {
	int i=blockDim.x*blockIdx.x+threadIdx.x;
	dest[i]=src[i];
}
int CUDAcopy1to2(void) {
	CUDAcopykernel<<<nCUDAbytes/blockSize/4,blockSize>>>((float *)CUDAdevice,(float *)CUDAdevice2);
	return 0;
}

int CUDAreadram(void) {
	int *arr=(int *)CUDAhost; int n=nCUDAbytes/sizeof(int);
	int sum=0;
	for (int i=0;i<n;i++) sum+=arr[i];
	return sum;
}
int CUDAwriteram(void) {
	int *arr=(int *)CUDAhost; int n=nCUDAbytes/sizeof(int);
	for (int i=0;i<n;i++) arr[i]=0;
	return 0;
}

void validateCUDAcopy(std::string type,int doAsync) {
	validate("cudaMemcpy HtoD "+type,CUDAh2d,CUDAswitchsize);
	validate("cudaMemcpy DtoH "+type,CUDAd2h,CUDAswitchsize);
	if (doAsync) {
		worktime=40.0e-6;
		validate("cudaMemcpyAsync HtoD "+type,CUDAh2dA,CUDAswitchsize);
		validate("cudaMemcpyAsync DtoH "+type,CUDAd2hA,CUDAswitchsize);
		
		/* Check how probe interval affects timing */
		CUDAasyncprint=1;
		for (nCUDAbytes=1024;nCUDAbytes<=max_size;nCUDAbytes*=16) 
		for (worktime=1.0e-6;worktime<=70.0e-6;worktime+=10.0e-6) {
			CUDAh2dA();
		}
		CUDAasyncprint=0;
	}
}

void validateCUDA(void) {
	
	
/* Memory management */
	validate("new_delete",CUDAhostnewdel,CUDAswitchsize);
	validate("cudaMalloc",CUDAdevmalloc,CUDAswitchsize);
	validate("cudaMallocHost",CUDAhostmalloc,CUDAswitchsize);
	
	cudaMalloc(&CUDAdevice,max_size);
	cudaMalloc(&CUDAdevice2,max_size);
	
	doCudaSync=1; /* else CUDA calls lie about completion time */
	CUDAhost=new char[max_size];
	validateCUDAcopy("unpinned",0);
	delete[] (char *)CUDAhost;
	
	cudaMallocHost(&CUDAhost,max_size);
	validateCUDAcopy("pinned",1);
	
	cudaFreeHost(CUDAhost);

/* Kernel execution */
	for (blockSize=1;blockSize<=512;blockSize*=2) {
		char name[100]; sprintf(name,"writeZeros,block=%03d",blockSize);
		int max_blocks=65535;
		int block_limit=max_blocks*blockSize*4;
		int max_bytes=std::min((int)max_size,block_limit);
		validate(name,CUDAbasicwrite,CUDAswitchsize,blockSize*4,max_bytes);
	}
	blockSize=128;
	validate("copydata,block=128",CUDAcopy1to2,CUDAswitchsize,blockSize*4);
	validate("cudaMemcpy DtoD",CUDAd2d,CUDAswitchsize);

/* Mapped memory */
	cudaDeviceProp deviceProp;
	cudaGetDeviceProperties(&deviceProp, 0);
#if CUDART_VERSION >= 2020
	if(deviceProp.canMapHostMemory) {
		validate("cudaHostAllocMapped",CUDAhostmappedmalloc,CUDAswitchsize);
		
		/* Test writecombined memory */
		cudaHostAlloc(&CUDAhost,max_size,cudaHostAllocMapped+cudaHostAllocWriteCombined);
		printf("Allocated host memory at %p\n",CUDAhost);
		CUDAerrorcheck("after cudaHostAlloc");
		
		void *devhost=0; /* host memory's device-mapped address */
		cudaHostGetDevicePointer(&devhost,CUDAhost,0);
		printf("Allocated device memory at %p\n",devhost);
		CUDAerrorcheck("after getDevicePointer");
		
		std::swap(devhost,CUDAdevice); /*hostmapped memory is read-side */
		validate("cudaGPUreadMap",CUDAcopy1to2,CUDAswitchsize,blockSize*4);
		std::swap(devhost,CUDAdevice); /* restore */
		
		std::swap(devhost,CUDAdevice2); /*hostmapped memory is write-side */
		validate("cudaGPUwriteMap",CUDAcopy1to2,CUDAswitchsize,blockSize*4);
		std::swap(devhost,CUDAdevice2); /* restore */
		
		validate("cudaCPUreadMapWC",CUDAreadram,CUDAswitchsize);
		validate("cudaCPUwriteMapWC",CUDAwriteram,CUDAswitchsize);
		
		cudaFreeHost(CUDAhost);
		
		/* Test non-writecombined memory */
		cudaHostAlloc(&CUDAhost,max_size,cudaHostAllocMapped);
		
		validate("cudaCPUreadMap",CUDAreadram,CUDAswitchsize);
		validate("cudaCPUwriteMap",CUDAwriteram,CUDAswitchsize);
		
		cudaFreeHost(CUDAhost);
	}
#endif
	doCudaSync=0;
	
	if (totpe>=2) {
		validate("cudaMPI_oneway",cudaMPIoneway,CUDAswitchsize);
		validate("cudaMPI_pingpong",cudaMPIpingpong,CUDAswitchsize);
		validate("cudaMPI_Ioneway",cudaMPI_Ioneway,CUDAswitchsize);
	}

	validate("cudaMPI_Bcast",cudaMPIbcast,CUDAswitchsize);
	validate("cudaMPI_Reduce",cudaMPIreduce,CUDAswitchsize,1,max_size/2);
	validate("cudaMPI_Allreduce",cudaMPIallreduce,CUDAswitchsize,1,max_size/2);
	
	debug("Freeing CUDAdevice memory...");
	cudaFree(CUDAdevice);
	cudaFree(CUDAdevice2);
}



/****************************** OpenGL performance model */

int GLwid=0,GLht=0;/* pixel size of buffers */
enum {GLwidmax=8192};
enum {GLhtmax=max_size/(GLwidmax*sizeof(float))};

gpu_env GLenv;
gpu_array *GLa, *GLb;
float *GLlandfill; /* unused area of CPU memory */

#if 1 /* Luminance data type for pixels (waaay more convenient for apps) */
enum {pixeltype=GL_LUMINANCE32F_ARB};
int GLreadfmt=GL_LUMINANCE, GLfloatperpix=1;
#else /* RGBA: more efficient on older hardware (equivalent on new stuff) */
enum {pixeltype=GL_RGBA32F_ARB};
int GLreadfmt=GL_RGBA, GLfloatperpix=4;
#endif

void GLerrorcheck(const char *where) {
        GLenum e=glGetError();
        if (0!=e) {
                printf("OpenGL error %d at %s\n",e,where);
                abort();
        }
}

int GLswitchsize(int newbytes) {
	GLerrorcheck("before switchsize");

	/* Avoid test-to-test contamination */
	glFinish();
	
	int wid=(newbytes+15)/(sizeof(float)*GLfloatperpix);
	if (wid<GLwidmax) { /* less than one row */
		GLwid=wid; GLht=1; 
	}
	else { /* compute whole rows */
		GLht=(wid+GLwidmax-1)/GLwidmax;
		if (GLht>GLhtmax) {GLht=GLhtmax;}
		GLwid=GLwidmax;
	}
	return GLwid*GLht*sizeof(float)*GLfloatperpix;
}
int GLswitchsize_cols(int newbytes) {
	GLerrorcheck("before switchsize");

	/* Avoid test-to-test contamination */
	glFinish();
	
	int ht=(newbytes+15)/(sizeof(float)*GLfloatperpix);
	if (ht<GLhtmax) { /* less than one col */
		GLht=ht; GLwid=1; 
	}
	else { /* compute whole cols */
		GLwid=(ht+GLhtmax-1)/GLhtmax;
		if (GLwid>GLwidmax) {GLwid=GLwidmax;}
		GLht=GLhtmax;
	}
	return GLwid*GLht*sizeof(float)*GLfloatperpix;
}

int GLtexbind(void)
{
	GLa->gpgpu_texture2D::bind();
	return 0;
}
int GLfbobind(void)
{
	GLa->gpgpu_framebuffer::bind();
	return 0;
}

int GLoneway(void)
{
	if (mype==sendPE) {
		GLa->gpgpu_framebuffer::bind();
		glMPI_Send(0,0, GLwid,GLht, pixeltype, recvPE,0,MPI_COMM_WORLD);
	}
	MPI_Status sts;
	if (mype==recvPE) {
		GLa->gpgpu_texture2D::bind();
		glMPI_Recv(0,0, GLwid,GLht, pixeltype, sendPE,0,MPI_COMM_WORLD,&sts); 
	}
	return 0;
}
int GLpingpong(void) {
	GLoneway(); std::swap(sendPE,recvPE); return 0;
}

int GLreadpixels(void) {
	GLa->gpgpu_framebuffer::bind();
	glReadPixels(0,0, GLwid,GLht, GLreadfmt,GL_FLOAT,GLlandfill);
	return 0;
}
int GLdrawpixels(void) {
	GLa->gpgpu_framebuffer::bind();
	glWindowPos2iARB(0,0);
	glDrawPixels(GLwid,GLht, GLreadfmt,GL_FLOAT,GLlandfill);
	return 0;
}
int GLtexsub(void) {
	GLa->gpgpu_texture2D::bind();
	glTexSubImage2D(GL_TEXTURE_2D,0,
		0,0,GLwid,GLht, GLreadfmt,GL_FLOAT,GLlandfill);
	return 0;
}

GLuint GLpbo=0;
int GLdata_pbo(void) {
	glBindBuffer(GL_PIXEL_UNPACK_BUFFER_ARB, GLpbo);
	glBufferData(GL_PIXEL_UNPACK_BUFFER_ARB, GLwid*GLht*GLfloatperpix*sizeof(float), GLlandfill, GL_STREAM_DRAW);
	glBindBuffer(GL_PIXEL_UNPACK_BUFFER_ARB, 0);
	return 0;
}
int GLtexsub_pbo(void) {
	glBindBuffer(GL_PIXEL_UNPACK_BUFFER_ARB, GLpbo);
	//glBufferData(GL_PIXEL_UNPACK_BUFFER_ARB, GLwid*GLht**GLfloatperpix*sizeof(float), GLlandfill, GL_STREAM_DRAW);
	glTexSubImage2D(GL_TEXTURE_2D,0,
		0,0,GLwid/GLfloatperpix,GLht, GLreadfmt,GL_FLOAT,0);
	glBindBuffer(GL_PIXEL_UNPACK_BUFFER_ARB, 0);
	return 0;
}

void validateGL(void) {
	doGLfinish=1;
	GLlandfill=new float[max_size/sizeof(float)];
	
	GLa=new gpu_array(GLenv,"A",GLwidmax,GLhtmax,0,pixeltype);
	GLb=new gpu_array(GLenv,"B",GLwidmax,GLhtmax,0,pixeltype);
	
	validate("glBindTexture",GLtexbind,GLswitchsize);
	validate("glBindFramebuffer",GLfbobind,GLswitchsize);
	validate("glReadPixels (rows)",GLreadpixels,GLswitchsize);
	validate("glReadPixels (cols)",GLreadpixels,GLswitchsize_cols);
	validate("glDrawPixels",GLdrawpixels,GLswitchsize);
	validate("glTexSubImage2D",GLtexsub,GLswitchsize);

	glGenBuffers(1, &GLpbo); /* need whole new pbo */
	glBindBuffer(GL_PIXEL_UNPACK_BUFFER_ARB, GLpbo);
	glBufferData(GL_PIXEL_UNPACK_BUFFER_ARB, max_size, GLlandfill, GL_STREAM_DRAW);
	glBindBuffer(GL_PIXEL_UNPACK_BUFFER_ARB, 0);
	validate("glBufferData (PBO)",GLdata_pbo,GLswitchsize);
	validate("glTexSubImage2D (PBO)",GLtexsub_pbo,GLswitchsize);
	glDeleteBuffers(1,&GLpbo); GLpbo=0;

	if (totpe>1) {
		validate("glMPI_Send (oneway)",GLoneway,GLswitchsize);
		validate("glMPI_Send (end to end)",GLpingpong,GLswitchsize);
	}
	
	delete GLa; delete GLb; delete[] GLlandfill; GLa=GLb=0; GLlandfill=0;
	doGLfinish=0;
}



/****************** Main **********************/

void call_finalize(void) {MPI_Finalize();}
int main(int argc,char *argv[]) {
	cudaSetDeviceFlags(cudaDeviceMapHost);
	
	int i;
	char *longMsg=(char *)malloc(sizeof(char)*LONGMSG);
	const char *curtest=""; /* Current test in progress */
	double starttime; /* Time test was started */
	int n; /* Number of iterations of test */
	MPI_Init(&argc,&argv);
	atexit(call_finalize); /*<- important to avoid weird errors! */
	MPI_Comm_rank(MPI_COMM_WORLD,&mype);
	MPI_Comm_size(MPI_COMM_WORLD,&totpe);
	//{char buf[100]; sprintf(buf,".%d",mype); log_suffix=buf;}
	if (argc>1) { /* change output dir */
		log_dir=argv[1];
	}
	if (argc>2) { /* change timing precision */
		step_pow=atof(argv[2]); 
		time_in_seconds_granularity=0.05; /* 50ms per test */
	}
	mkdir(log_dir.c_str(),0777);
	
	if (mype==0) test_clock();
	
	validateGL();
	
	validateCUDA();
	
	//if (mype==a || mype==b)
	//  printf("Starting processor %d of %d\n",mype,totpe);

	STARTTIME("sandbag (100)");
	for (i=0;i<1000;i++)
	{ 
	  sandbag(100);
	  n++;
	}
	ENDTIME

	validateMPI();
	
	
	return 0;
}
