/*
Test cudaMPI for correctness; basically checking if data is moving.
*/
#include "cudaMPI.h"
#include "cudaMPI.cpp" /* stupid, for linking ease */

#include <mpi.h> /* for MPI_ functions */
#include <cuda_runtime.h>

#include <stdio.h> /* for printf */
#include <stdlib.h> /* for atexit */
#include <math.h> /* for "cos" */
#include <sys/stat.h> /* for "mkdir" */
#include <vector>
#include <algorithm> /* for "pair" */

int mype=0,totpe=1;
 
int nCUDAbytes=1024; /* size of below buffers */
void *CUDAhost=0; /* host buffer */
void *CUDAdevice=0; /* device buffer 1*/
void *CUDAdevice2=0; /* device buffer 2*/

__global__ void CUDAwritefloat(float *dest,float v) {
	int i=blockDim.x*blockIdx.x+threadIdx.x;
	dest[i]=v;
}
int blockSize=32;
void setCuda(float v,void *dest=0) {
	if (dest==0) dest=CUDAdevice;
	CUDAwritefloat<<<nCUDAbytes/blockSize/4,blockSize>>>((float *)dest,v);
}
void printCuda(const char *where,void *what=0) {
	if (what==0) what=CUDAdevice;
	cudaMemcpy(CUDAhost,what,nCUDAbytes,cudaMemcpyDeviceToHost);
	float *f=(float *)CUDAhost;
	printf(" rank %d: CUDA %s = ",mype,where);
	for (int i=0;i<4;i++) printf("%.4f ",f[i]);
	printf("\n");
}

void CUDAerrorcheck(const char *where) {
	/* Check for any errors */
	cudaError_t e=cudaGetLastError();
	if (e!=cudaSuccess) {
		printf("CUDA error %d on pe %d %s: %s\n",
			e,mype,where,cudaGetErrorString(e));
		exit(1);
	}
}

int sendPE=0, recvPE=1;
void cudaMPIoneway(void) {
	int tag=rand()&0xffff;
	
	if (mype==sendPE) {
		setCuda(1.2345);
		cudaMPI_Send(CUDAdevice,8,MPI_BYTE, recvPE,tag,MPI_COMM_WORLD);
	}
	MPI_Status sts;
	if (mype==recvPE) {
		setCuda(0.0);
		cudaMPI_Recv(CUDAdevice,8,MPI_BYTE, sendPE,tag,MPI_COMM_WORLD,&sts);
	}
}
void cudaMPInonblock(int mode) {
	int tag=rand()&0xffff;
	std::vector<cudaMPI_Request> reqs; cudaMPI_Request req;
	
	if (mype==sendPE) {
		setCuda(1.2345,CUDAdevice2);
		cudaMPI_Isend(CUDAdevice2,8,MPI_BYTE, recvPE,tag,MPI_COMM_WORLD,&req);
		reqs.push_back(req);
	}
	if (mype==recvPE) {
		setCuda(0.0);
		cudaMPI_Irecv(CUDAdevice,8,MPI_BYTE, sendPE,tag,MPI_COMM_WORLD,&req);
		reqs.push_back(req);
	}
	
	int i,f;
	switch (mode) {
	case 0: cudaMPI_Waitall(reqs.size(),&reqs[0],0); break;
	case 1: for (i=0;i<reqs.size();i++) cudaMPI_Wait(&reqs[i],0); break;
	case 2: for (i=0;i<reqs.size();i++) {f=0; while (f==0) cudaMPI_Test(&reqs[i],&f,0); } break;
	};
}
void cudaMPIpingpong(void) {
	cudaMPIoneway(); std::swap(sendPE,recvPE);
}
void cudaMPIbcast(void) {
	int root=rand()%totpe; if (mype==0) printf("Root is %d\n",root);
	
	if (mype==root) setCuda(2.3456); else setCuda(10.0);
	cudaMPI_Bcast(CUDAdevice,3,MPI_FLOAT, root,MPI_COMM_WORLD);
}
void cudaMPIreduce(void) {
	int root=rand()%totpe; if (mype==0) printf("Root is %d\n",root);
	
	if (mype==root) setCuda(0.4567); else setCuda(3.0);
	cudaMPI_Reduce(CUDAdevice,CUDAdevice2,1,MPI_FLOAT,MPI_SUM,root,MPI_COMM_WORLD);
	printCuda("summed values",CUDAdevice2);
}

void testphase(const char *where) {
	printCuda("after test");
	MPI_Barrier(MPI_COMM_WORLD);
	
	setCuda(-1); setCuda(-1,CUDAdevice2);
	if (mype==0) printf("-----------------------------------\n%s\n",where);
}

void validateCUDA(void) {
/* Memory management */
	cudaMalloc(&CUDAdevice,nCUDAbytes);
	cudaMalloc(&CUDAdevice2,nCUDAbytes);
	cudaMallocHost(&CUDAhost,nCUDAbytes);
	
	testphase("one-way send test");
	cudaMPIoneway();
	
	testphase("pingpong send test");
	cudaMPIpingpong();
	cudaMPIpingpong();
	
	for (int mode=0;mode<=2;mode++) {
		testphase("nonblocking test");
		cudaMPInonblock(mode);
	}
	
	testphase("broadcast test");
	cudaMPIbcast();
	
	testphase("reduce test");
	cudaMPIreduce();
	
	
	testphase("tests complete");
	cudaFreeHost(CUDAhost);
	cudaFree(CUDAdevice);
	cudaFree(CUDAdevice2);
}


void call_finalize(void) {MPI_Finalize();}
int main(int argc,char *argv[]) {
	cudaSetDeviceFlags(cudaDeviceMapHost);
	
	MPI_Init(&argc,&argv);
	atexit(call_finalize); /*<- important to avoid weird errors! */
	MPI_Comm_rank(MPI_COMM_WORLD,&mype);
	MPI_Comm_size(MPI_COMM_WORLD,&totpe);
	
	if (totpe<2) sendPE=recvPE=0;
	
	srand(0); /* all MPI ranks will share the same random number stream */
	validateCUDA();
	
	return 0;
}
