/**
 cudaMPI: a message-passing interface for general-purpose graphics processing.
 Pass in GPU data, and this library will copy it out to the CPU
 and send it across the network using any MPI implementation.
 
 This library plays nicely with your own uses of MPI and CUDA.
 
 Dr. Orion Sky Lawlor, lawlor@alaska.edu, 2009-06-19 (Public Domain)
*/
#include "cudaMPI.h"
#include <cuda_runtime.h>
#include <vector> /* used to store streams for recycling */

/* Fatal error function */
static void cudaMPI_abort(const char *why) {
	printf("cudaMPI FATAL ERROR: %s\n",why);
	MPI_Finalize();
	abort();
}

/** A memory area (byte buffer) */
class cudaMPI_buf {
public:
	void *addr; /* start location of buffer */
	int len; /* length of buffer, in bytes */
	enum { initbuf=0, deadbuf, newbuf, pinbuf } type;
	cudaMPI_buf(void) { type=initbuf; addr=0; len=-1; }
};


/** All of cudaMPI's internal data is stored in this object, rather
than being scattered through a zillion global variables.  
Using one class helps make cudaMPI threadsafe. */
class cudaMPI_state_t {
public:
	/* CPU buffer recycling:
	  Rationale: cudaMallocHost provides pinned memory, which is 50% higher bandwidth,
	  but costs milliseconds to create.  So we allocate a bunch of pinned memory once,
	  and use that for everything.
	*/
	char *pinstart, *pinptr, *pinend;
	int pintot; // bytes of pinned memory available
	cudaMPI_state_t() :pinstart(0), pinptr(0), pinend(0), pintot(0) {}
	
	// Make sure there's at least this much pinned space available 
	//   uses stupid non-memory-reusing algorithm (FIXME: use heap)
	// Currently somewhat tolerable if in-use space (pintot) hits zero frequently.
	void addPinned(int len) {
		int alloc=(len*2+pintot+4095)&~4095; /* round up to next 4096-byte boundary */
		cudaMallocHost((void **)&pinstart,alloc);
		//printf("cudaMPI: allocated %d bytes of new pinned memory at %p\n",alloc,pinstart);
		pinptr=pinstart; pinend=pinstart+alloc;
		pintot=0;
	}
	
	
	/** Allocate a CPU buffer for this MPI data.
	    @param count Number of items needed
	    @param datatype MPI representation of item's types
	    @param mustBePinned If true, memory must be pinned on the host (more expensive,
	      but allows async copies).  If false, the default, memory may be pinned or unpinned.
	*/
	cudaMPI_buf allocateCPU(int count,MPI_Datatype datatype,bool mustBePinned=false)
	{
		cudaMPI_buf b; 
		int len=0;
		if (datatype==MPI_BYTE) len=count; 
		else if (datatype==MPI_FLOAT) len=count*sizeof(float);
		else if (datatype==MPI_DOUBLE) len=count*sizeof(double);
		else if (datatype==MPI_CHAR || datatype==MPI_UNSIGNED_CHAR || datatype==MPI_SIGNED_CHAR ) len=count*sizeof(char);
		else if (datatype==MPI_SHORT || datatype==MPI_UNSIGNED_SHORT) len=count*sizeof(short);
		else if (datatype==MPI_INT || datatype==MPI_UNSIGNED) len=count*sizeof(int);
		else if (datatype==MPI_LONG || datatype==MPI_UNSIGNED_LONG) len=count*sizeof(long);
		else cudaMPI_abort("Unrecognized datatype passed to allocateCPU, which only recognizes primitive contiguous datatypes.");
		
		
		b.len=len;
		if (true) { // mustBePinned) {  // use pinned memory
			if (pinptr+len>pinend) {
				addPinned(len); /* FIXME! very stupid; use a real heap algorithm */
			}
			pintot+=len;
			b.addr=pinptr;  pinptr+=len;
			b.type=cudaMPI_buf::pinbuf;
		} else { /* slower, but works: need not be pinned, just use new & delete */
			b.addr=new char[len];
			b.type=cudaMPI_buf::newbuf;
		}
		return b;
	}
	cudaMPI_buf toCPU(void *gpusrc,int count,MPI_Datatype datatype)
	{
		cudaMPI_buf b=allocateCPU(count,datatype);
		cudaMemcpy(b.addr,gpusrc,b.len,cudaMemcpyDeviceToHost);
		return b;
	}
	void fromCPUrelease(cudaMPI_buf &b,void *gpudest,int count,MPI_Datatype datatype)
	{
		cudaMemcpy(gpudest,b.addr,b.len,cudaMemcpyHostToDevice);
		releaseCPU(b);
	}
	void releaseCPU(cudaMPI_buf &b)
	{
		if (b.type==cudaMPI_buf::newbuf) {
			delete[] (char *)b.addr;
		} else if (b.type==cudaMPI_buf::pinbuf) {
		 	pintot-=b.len; // FIXME: use real heap
			if (pintot==0) pinptr=pinstart; /* reset buffer when empty */
		} else {
			cudaMPI_abort("Bad buffer type in releaseCPU");
		}
		b.addr=0; b.len=-1;
		b.type=cudaMPI_buf::deadbuf;
	}
	

	/* cudaStream recycling:
	  Rationale: cudaStreamCreate costs 40us, so we should save 
	  created streams for reuse.
	*/
	std::vector<cudaStream_t> streams;
	cudaStream_t getStream(void) {
		cudaStream_t s; 
		if (streams.size()==0u) {cudaStreamCreate(&s);} /* gotta make more */
		else {s=streams.back(); streams.pop_back();}
		return s;
	}
	void releaseStream(cudaStream_t s) {
		streams.push_back(s);
	}
};

/** Return our cudaMPI_state_t object, or create one if none exists */
static cudaMPI_state_t *cudaMPI_state_get(void) {
	static /* ...one big declaration */
#if defined(__GNUC__) && (__GNUC__>3 || (__GNUC__==3 && __GNUC_MINOR__>=3))
	__thread /* make cudaMPI_state_t thread-local */
#endif
	/* declaration continues... */ cudaMPI_state_t *s=0;
	
	if (s==0) s=new cudaMPI_state_t;
	return s;
}





/* Point-to-point communication */
CUDAMPI_ENTRY int cudaMPI_Send(void *gpubuf, int count, MPI_Datatype datatype, int dest, int tag, MPI_Comm comm)
{
	cudaMPI_state_t *s=cudaMPI_state_get();
	cudaMPI_buf b=s->toCPU(gpubuf,count,datatype);
	
	int r=MPI_Send(b.addr,count,datatype,dest,tag,comm);
	
	s->releaseCPU(b);
	return r;
}
CUDAMPI_ENTRY int cudaMPI_Recv(void *gpubuf, int count, MPI_Datatype datatype, int source, int tag, MPI_Comm comm, MPI_Status *status)
{
	cudaMPI_state_t *s=cudaMPI_state_get();
	cudaMPI_buf b=s->allocateCPU(count,datatype);
	
	int r=MPI_Recv(b.addr,count,datatype,source,tag,comm,status);
	
	s->fromCPUrelease(b,gpubuf,count,datatype);
	return r;
}

/* Collective communication */
CUDAMPI_ENTRY int cudaMPI_Bcast(void *gpubuf, int count, MPI_Datatype datatype, int root, MPI_Comm comm)
{
	cudaMPI_state_t *s=cudaMPI_state_get();
	cudaMPI_buf b; int mype; MPI_Comm_rank(comm,&mype);
	if (mype==root) b=s->toCPU(gpubuf,count,datatype); /* root sends */
	else b=s->allocateCPU(count,datatype); /* others recv */
	
	int r=MPI_Bcast(b.addr,count,datatype,root,comm);
	
	if (mype==root) s->releaseCPU(b);
	else s->fromCPUrelease(b,gpubuf,count,datatype);
	return r;
	
}
CUDAMPI_ENTRY int cudaMPI_Reduce(void *gpusendbuf, void *gpurecvbuf, int count, MPI_Datatype datatype, MPI_Op op, int root, MPI_Comm comm)
{
	cudaMPI_state_t *s=cudaMPI_state_get();
	cudaMPI_buf br,b=s->toCPU(gpusendbuf,count,datatype); /* everybody sends */
	int mype; MPI_Comm_rank(comm,&mype);
	if (mype==root) br=s->allocateCPU(count,datatype); /* root recvs */
	
	int r=MPI_Reduce(b.addr,br.addr,count,datatype,op,root,comm);
	
	if (mype==root) s->fromCPUrelease(br,gpurecvbuf,count,datatype);
	s->releaseCPU(b);
	return r;
}
CUDAMPI_ENTRY int cudaMPI_Allreduce(void *gpusendbuf, void *gpurecvbuf, int count, MPI_Datatype datatype, MPI_Op op, MPI_Comm comm)
{
	cudaMPI_state_t *s=cudaMPI_state_get();
	cudaMPI_buf b=s->toCPU(gpusendbuf,count,datatype); /* everybody sends */
	cudaMPI_buf br=s->allocateCPU(count,datatype); /* everybody recvs */
	
	int r=MPI_Allreduce(b.addr,br.addr,count,datatype,op,comm);
	
	s->fromCPUrelease(br,gpurecvbuf,count,datatype);
	s->releaseCPU(b);
	return r;
}

/**************************************************************/
/* Non-blocking communication */
struct cudaMPI_request_t {
	cudaMPI_state_t *s; /* our state (to avoid thread-specific lookups) */
	cudaMPI_buf b; /* CPU data buffer */
	cudaStream_t stream; /* CUDA stream object */
	MPI_Request req; /* underlying MPI request handle */
	MPI_Status sts; /* MPI status */
	
	/* stashed function parameters */
	void *gpubuf; int count; MPI_Datatype datatype; int proc,tag;MPI_Comm comm;
	
	/* Operations to perform: e.g., CSMD (Cuda stream, Send, MPI_Request, Done).
	  "op" is an index into the "ops" string. */
	char ops[8]; int op;
	
	cudaMPI_request_t(const char *ops_,void *gpubuf_, int count_, MPI_Datatype datatype_, int proc_, int tag_, MPI_Comm comm_);
	
	/* Run this request's state machine.  Returns true if we're now done, 
	   and should be removed from the list of active requests... */
	bool advance(bool canBlock);
	
	~cudaMPI_request_t();
};

cudaMPI_request_t::cudaMPI_request_t(const char *ops_,void *gpubuf_, int count_, MPI_Datatype datatype_, int proc_, int tag_, MPI_Comm comm_)
	:s(cudaMPI_state_get()), b(s->allocateCPU(count_,datatype_,true)), stream(s->getStream()), req(0), 
	gpubuf(gpubuf_),count(count_),datatype(datatype_),proc(proc_),tag(tag_),comm(comm_),
	op(0)
{
	if (strlen(ops_)>=sizeof(ops)) cudaMPI_abort("Bad list of operations while creating cudaMPI_Request");
	strcpy(ops,ops_); /* copy over operation list */
}

bool cudaMPI_request_t::advance(bool canBlock) {
	while (1) {
		switch (ops[op]) {
		case 'M': /* probe on a MPI request */
			if (canBlock) MPI_Wait(&req,&sts);
			else { 
				int flag=0; 
				MPI_Test(&req,&flag,&sts); 
				if (flag==0) return false; /* MPI isn't done yet */
			}
			break;
		case 'C': /* query CUDA stream */
			if (canBlock) cudaStreamSynchronize(stream);
			else { 
				if (cudaErrorNotReady==cudaStreamQuery(stream)) 
					return false; /* CUDA not done yet */
			}
			break;
		case 'S': /* MPI Send */
			MPI_Isend(b.addr,count,datatype,proc,tag,comm,&req);
			break;
		case 'f': /* data copy from CPU */
			cudaMemcpyAsync(gpubuf,b.addr,b.len,cudaMemcpyHostToDevice,stream);
			break;
		case 'D': /* done! */
			return true;
		default:
			cudaMPI_abort("Invalid operation found in cudaMPI_Request object");
		};
		op++; /* advance to next operation */
	}
}

cudaMPI_request_t::~cudaMPI_request_t()
{
	s->releaseStream(stream);
	s->releaseCPU(b);
	s=0; /* self-destruct mode, to detect use-after-delete errors */
}

CUDAMPI_ENTRY int cudaMPI_Irecv(void *gpubuf, int count, MPI_Datatype datatype, int proc, int tag, MPI_Comm comm, cudaMPI_Request *request)
{
	/* recv on CPU, wait for recv, copy to GPU, wait for copy, done. */
	cudaMPI_request_t *r=new cudaMPI_request_t("MfCD",gpubuf,count,datatype,proc,tag,comm);
	int ret=MPI_Irecv(r->b.addr,count,datatype,proc,tag,comm,&r->req);
	*request=r;
	return ret;
}
CUDAMPI_ENTRY int cudaMPI_Isend(void *gpubuf, int count, MPI_Datatype datatype, int proc, int tag, MPI_Comm comm, cudaMPI_Request *request)
{
	/* copy to CPU, wait for copy, send out, wait for send, done. */
	cudaMPI_request_t *r=new cudaMPI_request_t("CSMD",gpubuf,count,datatype,proc,tag,comm);
	cudaMemcpyAsync(r->b.addr,gpubuf,r->b.len,cudaMemcpyDeviceToHost,r->stream);
	*request=r;
	return MPI_SUCCESS;
}

CUDAMPI_ENTRY int cudaMPI_Test(cudaMPI_Request *request, int *flag, MPI_Status *status)
{
	cudaMPI_request_t *r=*request;
	if (r==NULL) *flag=1;
	else {
		if (r->advance(false)) { /* we're done! */
			*flag=1;
			if (status!=MPI_STATUS_IGNORE) *status=r->sts;
			delete r;
			*request=NULL;
		}
		else *flag=0; /* not done yet */
	}
	return MPI_SUCCESS;
}
CUDAMPI_ENTRY int cudaMPI_Wait(cudaMPI_Request *request, MPI_Status *status)
{
	cudaMPI_request_t *r=*request;
	if (r!=NULL) {
		while (!r->advance(true)) { /* not done yet */ }
		if (status!=MPI_STATUS_IGNORE) *status=r->sts;
		delete r;
		*request=NULL;
	}
	return MPI_SUCCESS;
}
CUDAMPI_ENTRY int cudaMPI_Waitall(int count, cudaMPI_Request *array_of_requests, MPI_Status *array_of_statuses)
{
	bool allDone=true;
	do {
		allDone=true; /* assume they're all done */
		for (int i=0;i<count;i++) {
			cudaMPI_request_t *r=array_of_requests[i];
			if (r!=NULL) {
				if (!r->advance(false)) { /* not done yet */
					allDone=false;
				}
				else { /* this one's now done */
					if (array_of_statuses && array_of_statuses!=MPI_STATUS_IGNORE) 
						array_of_statuses[i]=r->sts;
					delete r;
					array_of_requests[i]=NULL;
				}
			}
		}
	} while (!allDone);
	return MPI_SUCCESS;
}

