/**
 cudaMPI: a message-passing interface for general-purpose graphics processing.
 Pass in GPU data, and this library will copy it out to the CPU
 and send it across the network using any MPI implementation.
 
 This library plays nicely with your own uses of MPI and CUDA.
 
 To compile with both CUDA and OpenMPI, I use this line:
   nvcc -O3 -I/usr/lib/openmpi/include/ -lmpi -lmpi_cxx -Xlinker -R/usr/local/cuda/lib cudaMPI.cpp -c -o cudaMPI.o
 
 Dr. Orion Sky Lawlor, lawlor@alaska.edu, 2009-06-19 (Public Domain)
*/
#ifndef __CUDAMPI_H
#define __CUDAMPI_H

#include <mpi.h> /* underlying implementation needed for MPI_ datatypes */

#ifdef __cplusplus /* cudaMPI is written in C++, but can be called from plain C */
extern "C" {
#endif

#define CUDAMPI_ENTRY /* empty; possibly used eventually for a Windows DLL */

/**************** MPI Standard Functions ***************/
/* We don't touch these yet, but we might someday... */
typedef MPI_Status cudaMPI_Status;
typedef MPI_Datatype cudaMPI_Datatype;
typedef MPI_Comm cudaMPI_Comm;
#define cudaMPI_Init MPI_Init
#define cudaMPI_Finalize MPI_Finalize

/* Point-to-point communication */
CUDAMPI_ENTRY int cudaMPI_Send(void *gpubuf, int count, MPI_Datatype datatype, int dest, int tag, MPI_Comm comm);
CUDAMPI_ENTRY int cudaMPI_Recv(void *gpubuf, int count, MPI_Datatype datatype, int source, int tag, MPI_Comm comm, MPI_Status *status);

/* Collective communication */
CUDAMPI_ENTRY int cudaMPI_Bcast(void *gpubuf, int count, MPI_Datatype datatype, int root, MPI_Comm comm);
CUDAMPI_ENTRY int cudaMPI_Reduce(void *gpusendbuf, void *gpurecvbuf, int count, MPI_Datatype datatype, MPI_Op op, int root, MPI_Comm comm);
CUDAMPI_ENTRY int cudaMPI_Allreduce(void *gpusendbuf, void *gpurecvbuf, int count, MPI_Datatype datatype, MPI_Op op, MPI_Comm comm);

/* Non-blocking communication */
typedef struct cudaMPI_request_t *cudaMPI_Request;
CUDAMPI_ENTRY int cudaMPI_Irecv(void *gpubuf, int count, MPI_Datatype datatype, int source, int tag, MPI_Comm comm, cudaMPI_Request *request);
CUDAMPI_ENTRY int cudaMPI_Isend(void *gpubuf, int count, MPI_Datatype datatype, int dest, int tag, MPI_Comm comm, cudaMPI_Request *request);
CUDAMPI_ENTRY int cudaMPI_Test(cudaMPI_Request *request, int *flag, MPI_Status *status);
CUDAMPI_ENTRY int cudaMPI_Wait(cudaMPI_Request *request, MPI_Status *status);
CUDAMPI_ENTRY int cudaMPI_Waitall(int count, cudaMPI_Request *array_of_requests, MPI_Status *array_of_statuses);

/****************** Non-MPI functions ****************/
/** This object describes a region of memory accessible to both CPU and GPU. */
typedef struct cudaMPI_Mapped_buffer {
	void *cpu_base; /* start of buffer in CPU address space */
	void *gpu_base; /* start of buffer in GPU address space */
	int bytes; /* length of buffer, in bytes */
} cudaMPI_Mapped_buffer;

#ifdef __cplusplus
};
#endif
#endif
