/**
 glMPI: a message-passing interface for general-purpose graphics processing.
 Communication interface written in terms of GPU framebuffer objects.
 
 Dr. Orion Sky Lawlor, lawlor@alaska.edu, 2009-06-24 (Public Domain)
*/
#ifndef __glMPI_H
#define __glMPI_H

/* You MUST include OpenGL and extensions before this header! */

#include <mpi.h> /* underlying implementation needed for MPI_ datatypes */
#include <stdio.h>
#include <stdlib.h>

/** This class represents the size of OpenGL pixel data in CPU memory. */
class glMPI_BufSize {
public:
	int size; /* total buffer size, in bytes */
	GLenum glformat; /* OpenGL's format, like GL_RGBA */
	GLenum gltype; /* OpenGL data type, like GL_FLOAT */
	int mpicount; /* number of MPI data items */
	MPI_Datatype mpitype; /* MPI data type, usually MPI_FLOAT */
	glMPI_BufSize(int width,int height,GLenum internalFormat) 
	{
		mpicount=width*height;
		mpitype=MPI_FLOAT;
		gltype=GL_FLOAT;
		int bytesper=sizeof(float); /* total buffer size: bytesper * mpicount */
		switch(internalFormat) {
		case GL_RGBA32F_ARB:
			glformat=GL_RGBA;
			mpicount*=4; /* 4 channels */
			break;
		case GL_LUMINANCE32F_ARB:
			glformat=GL_LUMINANCE;
			break;
		default:
			printf("Error! Unknown glMPI format %d!\n",internalFormat);
			abort();
		};
		size=bytesper*mpicount;
	}
};

/** This class *stores* OpenGL pixel data in CPU memory. */
class glMPI_Buf : public glMPI_BufSize {
public:
	void *data; /* underlying CPU-side data */
	glMPI_Buf(int width,int height,GLenum internalFormat) 
		:glMPI_BufSize(width,height,internalFormat)
	{
		storage.resize(size);
		data=&storage[0];
	}
	/* Silly debugging function */
	void show(const char *where) {
		float *f=(float *)data;
		printf("%s: %.2f %.2f %.2f %.2f\n",
			where, f[0],f[mpicount*1/3],f[mpicount*2/3],f[mpicount-1]
		);
	}
private:
	std::vector<char> storage; /* underlying CPU-side data */
};

/**
  glMPI_Send: copy rendered OpenGL pixels out onto the network.
  You MUST have bound the appropriate rendering target before calling here!
  (e.g., glBindFramebufferEXT(GL_FRAMEBUFFER_EXT or GL_READ_FRAMEBUFFER_EXT,myFBO); )
  
  Pixels are read using glReadPixels.
	@param x   X coordinate of first pixel to copy out.
	@param y   Y coordinate of first pixel to copy out.
	@param width  X size of pixel area to copy out.
	@param height Y size of pixel area to copy out.
	@param internalFormat GL-style pixel data type, one of GL_RGBA32F_ARB or GL_LUMINANCE32F_ARB.
	@param dest  Destination MPI process rank.
	@param tag   User-defined integer identifier.
	@param comm  MPI communicator, such as MPI_COMM_WORLD.
*/
inline int glMPI_Send(int x, int y, int width,int height, GLenum internalFormat,
	int dest, int tag, MPI_Comm comm)
{
	glMPI_Buf buf(width,height,internalFormat);
	glReadPixels(x,y,width,height,
		buf.glformat,buf.gltype,buf.data);
	//buf.show("glMPI_Send");
	return MPI_Send(buf.data,buf.mpicount,buf.mpitype,
		dest,tag,comm);
}

/**
  glMPI_Recv: copy OpenGL pixels off the network onto the rendering target.
  You MUST have bound your target texture (to receive pixels) before calling here!
  (e.g., glBindTexture(GL_TEXTURE_2D,myTex); )
  
  Pixels are written using glDrawPixels.
	@param x   X coordinate of first texture pixel to receive.
	@param y   Y coordinate of first texture pixel to receive.
	@param width  X size of pixel area to receive.
	@param height Y size of pixel area to receive.
	@param internalFormat GL-style pixel data type, one of GL_RGBA32F_ARB or GL_LUMINANCE32F_ARB.
	@param dest  Destination MPI process rank.
	@param tag   User-defined integer identifier.
	@param comm  MPI communicator, such as MPI_COMM_WORLD.
*/
inline int glMPI_Recv(int x, int y, int width,int height, GLenum internalFormat,
	int source, int tag, MPI_Comm comm, MPI_Status *status)
{
#if 1 /* pixel buffer object */
	glMPI_BufSize buf(width,height,internalFormat);
	static GLuint pbo=0; static int pbosize=0;
	if (buf.size>pbosize) 
	{ /* need to enlarge pbo */
		if (pbo==0) glGenBuffers(1, &pbo); /* need whole new pbo */
        	glBindBuffer(GL_PIXEL_UNPACK_BUFFER_ARB, pbo);
		pbosize=buf.size;
		glBufferData(GL_PIXEL_UNPACK_BUFFER_ARB, pbosize, NULL, GL_STREAM_DRAW);
	}
	/* Receive directly into OpenGL PBO buffer */
        glBindBuffer(GL_PIXEL_UNPACK_BUFFER_ARB, pbo);
	void *bufdata = glMapBuffer(GL_PIXEL_UNPACK_BUFFER_ARB,GL_WRITE_ONLY);
	int r=MPI_Recv(bufdata,buf.mpicount,buf.mpitype,
		source,tag,comm,status);
	glUnmapBuffer(GL_PIXEL_UNPACK_BUFFER_ARB);
	
	#if 1/* upload PBO data to a texture using glTexSubImage: over 100 MB/s! */
		glTexSubImage2D(GL_TEXTURE_2D,0,
			x,y,width,height,
			buf.glformat,buf.gltype,0);
	#else /* Draw the PBO data to a framebuffer object with drawPixels: pathetic 10 MB/s */
		glUseProgramObjectARB(0); /* don't run a program on the drawn pixels */
		glWindowPos2iARB(x,y); /* <- ARB_window_pos extension; sets glRasterPos to pixels */
		glDrawPixels(width,height, buf.glformat,buf.gltype,0); /* <-"0" means "from pbo" */
	#endif
	
        glBindBuffer(GL_PIXEL_UNPACK_BUFFER_ARB, 0);
#endif

#if 0 /* texSubImage: horrific alpha cost proportional to image size */
	glMPI_BufSize buf(width,height,internalFormat);
	int r=MPI_Recv(buf.data,buf.mpicount,buf.mpitype,
		source,tag,comm,status);
/* Absurdly, atrociously slow: for some reason, scales like *total* image size! */
	glTexSubImage2D(GL_TEXTURE_2D,0,
		x,y,width,height,
		buf.glformat,buf.gltype,buf.data);
#endif
#if 0 /* drawPixels: pathetic 10 MB/s */
	glMPI_BufSize buf(width,height,internalFormat);
	int r=MPI_Recv(buf.data,buf.mpicount,buf.mpitype,
		source,tag,comm,status);
	//buf.show("glMPI_Recv");
	glColor4f(1.0f,1.0f,1.0f,1.0f); /* else received data is colorized? */
	glUseProgramObjectARB(0); /* don't run a program on the drawn pixels */
	glWindowPos2iARB(x,y); /* <- ARB_window_pos extension; sets glRasterPos to pixels */
	
	glDrawPixels(width,height, buf.glformat,buf.gltype,buf.data);
#endif
	return r;
}


#endif
