/**
Trivial OpenGL program with a stencil-on-grid ("Jacobi")
style computation.  This is useful for computational
benchmarking (MFLOPS), communication benchmarking
(Mbytes/sec), synchronization tests (barriers/sec), 
and MPI/AMPI comparison.


PROCESSING:
The grid is a 2D array, with DEPTH rows and WIDTH
columns on each processor (i.e., this is a scaling problem).
This means each iteration performs 4*DEPTH*WIDTH single-precision
floating-point operations.

COMMUNICATION:
A processor's neighbors are above and below him, so 
the ghost regions are simply rows--this makes communication
simpler, since the data can be sent and received directly 
out of the data array, with no intermediate packing arrays.
Each message is of size WIDTH floats, and there are two 
messages, so a total of 8*WIDTH bytes are communicated
at each iteration.

computation/communication ratio: DEPTH/2 flops per communicated byte.


COMMAND-LINE USAGE:

c1000: Sets WIDTH (here, to 1000), which controls the amount of 
       communication per step.  
       Does not affect the computation/communication ratio.  
       The default is 100, for 800 bytes per step.
p1000: Sets DEPTH (here, to 1000), which controls the amount of 
       computation per step.  This is the computation/communication ratio.
       The default is 1000, for 500 flops per communicated byte.
n1000: Sets the number of iterations (here, to 1000).
       The default is 1000; of which the first 100 are not timed
       (to ignore startup transients).

B+/B-: Turn a per-iteration global barrier (MPI_Barrier) on and off.
       The default is for the global barrier to be off.
C+/C-: Turns communication (neighbor messages) on and off.  
       Communication is on by default.
P+/P-: Turns processing (flops) on and off.  
       Processing is on by default.
       
Examples:

Scaled problem run (default sizes):
    mpirun -np1 ./pgm
    mpirun -np2 ./pgm 
    mpirun -np4 ./pgm
    ...

Change communication size for fixed computation/communication ratio:
    mpirun -np4 ./pgm +p4 p10 c10
    mpirun -np4 ./pgm +p4 p10 c100
    mpirun -np4 ./pgm +p4 p10 c1000
    mpirun -np4 ./pgm +p4 p10 c10000
    mpirun -np4 ./pgm +p4 p10 c100000

Change communication size for fixed computation size:
    mpirun -np4 ./pgm +p4 p1000000 c10
    mpirun -np4 ./pgm +p4 p100000 c100
    mpirun -np4 ./pgm +p4 p10000 c1000
    mpirun -np4 ./pgm +p4 p1000 c10000
    mpirun -np4 ./pgm +p4 p100 c100000

Barrier rate:
    mpirun -np4 ./pgm +p4 B+ C- P-

Barrier rate vs. compute delay (nonlinear!):
    mpirun -np4 ./pgm +p4 B+ C- p1
    mpirun -np4 ./pgm +p4 B+ C- p10
    mpirun -np4 ./pgm +p4 B+ C- p100
    mpirun -np4 ./pgm +p4 B+ C- p1000
    mpirun -np4 ./pgm +p4 B+ C- p10000

Orion Sky Lawlor, olawlor@acm.org, 2009-06-20 (Public Domain)
*/
#include <stdio.h>
#include <stdlib.h>

#include "ogl/gpgpu.h"
#include "ogl/glsl.cpp"

#include "mpi.h"
#include "glMPI.h"

void glErrCheck(const char *where) {
	GLenum e=glGetError();
	if (0!=e) {
		printf("OpenGL error %d at %s\n",e,where);
		abort();
	}
}

/* Data type for pixels: */
enum {pixeltype=GL_LUMINANCE32F_ARB};

class chunk {
	int width; /* communication is 2*sizeof(float)*WIDTH bytes */
	int depth; /* computation is 2 DEPTH flops per communicated float 
	              (4 flops per point, DEPTH points per column, 
		       2 communicated floats per column) */
	/**
	  (Host) data array is depth rows (plus two ghost rows on top and bottom)
	  by width columns (plus two boundary condition columns left and right)
	*/
	float *data;
	
	/** GPU data arrays */
	gpu_env e;
	gpu_array *SRC,*DEST;
public:
	chunk(int w,int d) 
		:width(w), depth(d)
	{
		data=new float[(d+2)*(w+2)];
	}
	
	/* Get CPU data: */
	float at(int x,int y) const {return data[y*(width+2)+x];}
	float &at(int x,int y) {return data[y*(width+2)+x];}
	
	/* Bind GPU arrays for reads and writes: */
	void bind(void) {
		SRC->gpgpu_texture2D::bind();
		SRC->gpgpu_framebuffer::bind(); // for sends and receives
	}

	int getWidth(void) const {return width;}
	int getDepth(void) const {return depth;}
	
	/**
	 Initialize this grid.
	*/
	void initialize(int offX,int offY) {
		for (int y=0;y<depth+2;y++)
		for (int x=0;x<width+2;x++) {
			int gx=x+offX, gy=y+offY; /* global location */
			int cx=gx-850, cy=gy-800; /* circle-centric coords */
			int cr=800;
			int radius2=cx*cx+cy*cy;
			if (radius2<cr*cr) at(x,y)=1.0; /* inside circle, bright */
			else at(x,y)=0.0+0.0001*gx+0.0001*gy; /* outside circle, dim */
		}
		
		/* Copy grid to GPU */
		glErrCheck("before creating grids");
		SRC=new gpu_array(e,"SRC",(width+2),(depth+2),data,pixeltype);
		DEST=new gpu_array(e,"DEST",(width+2),(depth+2),data,pixeltype);
		glErrCheck("after creating grids");
	}
	
	/** Dump our grid */
	void save(const char *fname) {
		FILE *file=fopen(fname,"wb");
		/* Copy grid to GPU */
		SRC->read(data,(width+2),(depth+2));
		fprintf(file,"P5\n%d %d\n255\n",width+2,depth+2);
		for (int y=0;y<depth+2;y++)
		for (int x=0;x<width+2;x++) {
			float f=at(x,y);
			unsigned char c=(unsigned char)(255.0*f);
			if (f<0) c=0; if (f>1) c=255;
			fputc(c,file);
		}
		fclose(file);
	}
	
	/**
	 Do one sweep over this grid.  Computation takes 4 flops per point.
	*/
	void compute(void) {
		GPU_RUN(*DEST,
			" float l=texture2D(SRC,location+vec2(-SRCScale.x,0.0)).r;\n"
			" float r=texture2D(SRC,location+vec2(+SRCScale.x,0.0)).r;\n"
			" float t=texture2D(SRC,location+vec2(0.0,-SRCScale.y)).r;\n"
			" float b=texture2D(SRC,location+vec2(0.0,+SRCScale.y)).r;\n"
			" gl_FragColor=0.25*(l+r+t+b); \n"
		)
		SRC->swapwith(*DEST);
	}
};

void call_finalize(void) {MPI_Finalize();}
int main(int argc, char** argv)
{
  int thisIndex=0,nblocks=1;
  MPI_Init(&argc, &argv);
  atexit(call_finalize); /*<- important to avoid weird errors! */
  MPI_Comm comm=MPI_COMM_WORLD;
  MPI_Comm_rank(comm, &thisIndex);
  MPI_Comm_size(comm, &nblocks);
  int niter = 1000;
  
  bool doB=false; // Barrier
  bool doC=true; // Communication
  bool doP=true; // Processing
  int WIDTH=100;
  int DEPTH=1000;
  for (int iarg=1;iarg<argc;iarg++) {
    switch (argv[iarg][0]) {
    case 'B': doB=argv[iarg][1]=='+'; break;
    case 'C': doC=argv[iarg][1]=='+'; break;
    case 'P': doP=argv[iarg][1]=='+'; break;
    case 'c': WIDTH=atoi(&argv[iarg][1]); break;
    case 'p': DEPTH=atoi(&argv[iarg][1]); break;
    case 'n': niter=atoi(&argv[iarg][1]); break;
    default:
      printf("Unrecognized argument '%s'\n",argv[iarg]);
    };
  }
  enum {BLOCKDIM=4};
  WIDTH=((WIDTH+2+BLOCKDIM-1)&~(BLOCKDIM-1))-2; /* round width+2 to next multiple of BLOCKDIM */
  DEPTH=((DEPTH+2+BLOCKDIM-1)&~(BLOCKDIM-1))-2; /* round up to next multiple */
  
  int perC=2*sizeof(float)*WIDTH; // Bytes communicated per step
  int perP=4*WIDTH*DEPTH; // Floating point operations per step
  if (thisIndex==0) {
    printf("Per-step barrier %s\n",doB?"enabled":"disabled");
    if (doC) printf("Communication enabled: %d bytes/step\n",perC);
    else printf("Communication disabled.\n");
    if (doP) printf("Processing enabled: %dx%d, %d flops/step\n",WIDTH,DEPTH,perP);
    printf("Running for %d steps...\n",niter);
  }
  
  chunk *cp = new chunk(WIDTH,DEPTH);

  int xp = (thisIndex+1)%nblocks;
  int xm = (thisIndex-1+nblocks)%nblocks;
  
  cp->initialize(0,thisIndex*DEPTH);
  
  MPI_Barrier(comm);
  
  // Timing collection: total time spent in each task:
  double s;
  double tB=0;
  double tC=0;
  double tP=0;
  int firstiter=niter/10; // First iteration to time (avoids startup weirdness)
  
  for(int iter=0; iter<niter; iter++) {
    if (doB) {
      s=MPI_Wtime();
      MPI_Barrier(comm);
      if (iter>=firstiter) tB+=MPI_Wtime()-s;
    }
    
    if (doC) {
      s=MPI_Wtime();
      cp->bind();
      glMPI_Send(1,1,	   WIDTH,1, pixeltype, xm, 4440, comm);
      glMPI_Send(1,DEPTH,  WIDTH,1, pixeltype, xp, 4441, comm);
      
      glMPI_Recv(1,DEPTH+1,WIDTH,1, pixeltype, xp, 4440, comm, 0);
      glMPI_Recv(1,0,      WIDTH,1, pixeltype, xm, 4441, comm, 0);
      if (iter>=firstiter) tC+=MPI_Wtime()-s;
    }
    
    if (doP) {
      s=MPI_Wtime();
      cp->compute();
      glFinish(); /* for accurate timings */
      if (iter>=firstiter) tP+=MPI_Wtime()-s;
    }
  }
  niter-=firstiter;
  
  if (thisIndex==0) {
    tB/=niter; tC/=niter; tP/=niter;
    printf("Computation complete. Time per iteration: \n");
    if (doB) printf("  %.9f seconds per barrier\n", tB);
    if (doC) printf("  %.9f seconds per communication (%.2f MB/s)\n",tC,1.0e-6*perC/tC);
    if (doP) printf("  %.9f seconds per processing (%.2f Mflops)\n",tP,1.0e-6*perP/tP);
    if (doC) printf("  %.1f%% parallel efficiency.\n",100.0*tP/(tP+tC+tB));
    printf("BOTTOMLINE %d %d %d  %d  %.8f %.8f %.8f  %d %d %d\n",
    	nblocks,WIDTH,DEPTH, niter, tB,tC,tP, 1,perC,perP);
  }
  
  /* Save output files */
  char buf[100]; sprintf(buf,"out_%d_of_%d.ppm",thisIndex,nblocks);
  cp->save(buf);
}
